from enum import Enum 


class NodeKind(Enum):
    UnexposedDecl = 1
    StructDecl = 2
    UnionDecl = 3
    ClassDecl = 4
    EnumDecl = 5
    FieldDecl = 6
    EnumConstantDecl = 7
    FunctionDecl = 8
    VarDecl = 9
    ParmDecl = 10
    ObjCInterfaceDecl = 11
    ObjCCategoryDecl = 12
    ObjCProtocolDecl = 13
    ObjCPropertyDecl = 14
    ObjCIvarDecl = 15
    ObjCInstanceMethodDecl = 16
    ObjCClassMethodDecl = 17
    ObjCImplementationDecl = 18
    ObjCCategoryImplDecl = 19
    TypedefDecl = 20
    CXXMethod = 21
    Namespace = 22
    LinkageSpec = 23
    Constructor = 24
    Destructor = 25
    ConversionFunction = 26
    TemplateTypeParameter = 27
    NonTypeTemplateParameter = 28
    TemplateTemplateParameter = 29
    FunctionTemplate = 30
    ClassTemplate = 31
    ClassTemplatePartialSpecialization = 32
    NamespaceAlias = 33
    UsingDirective = 34
    UsingDeclaration = 35
    TypeAliasDecl = 36
    ObjCSynthesizeDecl = 37
    ObjCDynamicDecl = 38
    CXXAccessSpecifier = 39
    FirstRef = 40
    ObjCSuperClassRef = 40
    ObjCProtocolRef = 41
    ObjCClassRef = 42
    TypeRef = 43
    CXXBaseSpecifier = 44
    TemplateRef = 45
    NamespaceRef = 46
    MemberRef = 47
    LabelRef = 48
    OverloadedDeclRef = 49
    VariableRef = 50
    FirstInvalid = 70
    InvalidFile = 70
    NoDeclFound = 71
    NotImplemented = 72
    InvalidCode = 73
    FirstExpr = 100
    UnexposedExpr = 100
    DeclRefExpr = 101
    MemberRefExpr = 102
    CallExpr = 103
    ObjCMessageExpr = 104
    BlockExpr = 105
    IntegerLiteral = 106
    FloatingLiteral = 107
    ImaginaryLiteral = 108
    StringLiteral = 109
    CharacterLiteral = 110
    ParenExpr = 111
    UnaryOperator = 112
    ArraySubscriptExpr = 113
    BinaryOperator = 114
    CompoundAssignOperator = 115
    ConditionalOperator = 116
    CStyleCastExpr = 117
    CompoundLiteralExpr = 118
    InitListExpr = 119
    AddrLabelExpr = 120
    StmtExpr = 121
    GenericSelectionExpr = 122
    GNUNullExpr = 123
    CXXStaticCastExpr = 124
    CXXDynamicCastExpr = 125
    CXXReinterpretCastExpr = 126
    CXXConstCastExpr = 127
    CXXFunctionalCastExpr = 128
    CXXAddrspaceCastExpr = 129
    CXXTypeidExpr = 130
    CXXBoolLiteralExpr = 131
    CXXNullPtrLiteralExpr = 132
    CXXThisExpr = 133
    CXXThrowExpr = 134
    CXXNewExpr = 135
    CXXDeleteExpr = 136
    UnaryExpr = 137
    ObjCStringLiteral = 138
    ObjCEncodeExpr = 139
    ObjCSelectorExpr = 140
    ObjCProtocolExpr = 141
    ObjCBridgedCastExpr = 142
    PackExpansionExpr = 143
    SizeOfPackExpr = 144
    LambdaExpr = 145
    ObjCBoolLiteralExpr = 146
    ObjCSelfExpr = 147
    OMPArraySectionExpr = 148
    ObjCAvailabilityCheckExpr = 149
    FixedPointLiteral = 150
    OMPArrayShapingExpr = 151
    OMPIteratorExpr = 152
    FirstStmt = 200
    UnexposedStmt = 200
    LabelStmt = 201
    CompoundStmt = 202
    CaseStmt = 203
    DefaultStmt = 204
    IfStmt = 205
    SwitchStmt = 206
    WhileStmt = 207
    DoStmt = 208
    ForStmt = 209
    GotoStmt = 210
    IndirectGotoStmt = 211
    ContinueStmt = 212
    BreakStmt = 213
    ReturnStmt = 214
    GCCAsmStmt = 215
    ObjCAtTryStmt = 216
    ObjCAtCatchStmt = 217
    ObjCAtFinallyStmt = 218
    ObjCAtThrowStmt = 219
    ObjCAtSynchronizedStmt = 220
    ObjCAutoreleasePoolStmt = 221
    ObjCForCollectionStmt = 222
    CXXCatchStmt = 223
    CXXTryStmt = 224
    CXXForRangeStmt = 225
    SEHTryStmt = 226
    SEHExceptStmt = 227
    SEHFinallyStmt = 228
    MSAsmStmt = 229
    NullStmt = 230
    DeclStmt = 231
    OMPParallelDirective = 232
    OMPSimdDirective = 233
    OMPForDirective = 234
    OMPSectionsDirective = 235
    OMPSectionDirective = 236
    OMPSingleDirective = 237
    OMPParallelForDirective = 238
    OMPParallelSectionsDirective = 239
    OMPTaskDirective = 240
    OMPMasterDirective = 241
    OMPCriticalDirective = 242
    OMPTaskyieldDirective = 243
    OMPBarrierDirective = 244
    OMPTaskwaitDirective = 245
    OMPFlushDirective = 246
    SEHLeaveStmt = 247
    OMPOrderedDirective = 248
    OMPAtomicDirective = 249
    OMPForSimdDirective = 250
    OMPParallelForSimdDirective = 251
    OMPTargetDirective = 252
    OMPTeamsDirective = 253
    OMPTaskgroupDirective = 254
    OMPCancellationPointDirective = 255
    OMPCancelDirective = 256
    OMPTargetDataDirective = 257
    OMPTaskLoopDirective = 258
    OMPTaskLoopSimdDirective = 259
    OMPDistributeDirective = 260
    OMPTargetEnterDataDirective = 261
    OMPTargetExitDataDirective = 262
    OMPTargetParallelDirective = 263
    OMPTargetParallelForDirective = 264
    OMPTargetUpdateDirective = 265
    OMPDistributeParallelForDirective = 266
    OMPDistributeParallelForSimdDirective = 267
    OMPDistributeSimdDirective = 268
    OMPTargetParallelForSimdDirective = 269
    OMPTargetSimdDirective = 270
    OMPTeamsDistributeDirective = 271
    OMPTeamsDistributeSimdDirective = 272
    OMPTeamsDistributeParallelForSimdDirective = 273
    OMPTeamsDistributeParallelForDirective = 274
    OMPTargetTeamsDirective = 275
    OMPTargetTeamsDistributeDirective = 276
    OMPTargetTeamsDistributeParallelForDirective = 277
    OMPTargetTeamsDistributeParallelForSimdDirective = 278
    OMPTargetTeamsDistributeSimdDirective = 279
    BuiltinBitCastExpr = 280
    OMPMasterTaskLoopDirective = 281
    OMPParallelMasterTaskLoopDirective = 282
    OMPMasterTaskLoopSimdDirective = 283
    OMPParallelMasterTaskLoopSimdDirective = 284
    OMPParallelMasterDirective = 285
    OMPDepobjDirective = 286
    OMPScanDirective = 287
    TranslationUnit = 300
    FirstAttr = 400
    UnexposedAttr = 400
    IBActionAttr = 401
    IBOutletAttr = 402
    IBOutletCollectionAttr = 403
    CXXFinalAttr = 404
    CXXOverrideAttr = 405
    AnnotateAttr = 406
    AsmLabelAttr = 407
    PackedAttr = 408
    PureAttr = 409
    ConstAttr = 410
    NoDuplicateAttr = 411
    CUDAConstantAttr = 412
    CUDADeviceAttr = 413
    CUDAGlobalAttr = 414
    CUDAHostAttr = 415
    CUDASharedAttr = 416
    VisibilityAttr = 417
    DLLExport = 418
    DLLImport = 419
    NSReturnsRetained = 420
    NSReturnsNotRetained = 421
    NSReturnsAutoreleased = 422
    NSConsumesSelf = 423
    NSConsumed = 424
    ObjCException = 425
    ObjCNSObject = 426
    ObjCIndependentClass = 427
    ObjCPreciseLifetime = 428
    ObjCReturnsInnerPointer = 429
    ObjCRequiresSuper = 430
    ObjCRootClass = 431
    ObjCSubclassingRestricted = 432
    ObjCExplicitProtocolImpl = 433
    ObjCDesignatedInitializer = 434
    ObjCRuntimeVisible = 435
    ObjCBoxable = 436
    FlagEnum = 437
    ConvergentAttr = 438
    WarnUnusedAttr = 439
    WarnUnusedResultAttr = 440
    AlignedAttr = 441
    PreprocessingDirective = 500
    MacroDefinition = 501
    MacroExpansion = 502
    InclusionDirective = 503
    ModuleImportDecl = 600
    TypeAliasTemplateDecl = 601
    StaticAssert = 602
    FriendDecl = 603
    OverloadCandidate = 700
    FirstDecl = 1
    LastDecl = 39
    LastRef = 50
    LastInvalid = 73
    LastExpr = 152
    AsmStmt = 215
    LastStmt = 287
    LastAttr = 441
    MacroInstantiation = 502
    FirstPreprocessing = 500
    LastPreprocessing = 503
    FirstExtraDecl = 600
    LastExtraDecl = 603
    
    ClassTemplateSpecialization = 1001
    DecompositionDecl = 1002
    BindingDecl = 1003


from pathlib import Path

from ccimport import compat


def test_build():
    from myclang import cext, cindex


def test_parse():
    from myclang import cindex
    from myclang.cindex import TranslationUnit
    path = Path(__file__).parent / "code.cc"
    index = cindex.Index.create()
    cxxflags = [
        "-std=c++17",
    ]
    tu = index.parse(path,
                     args=[*cxxflags],
                     options=TranslationUnit.PARSE_NONE)


if __name__ == "__main__":
    test_parse()



# https://ziglang.org/deps/llvm%2bclang%2blld-11.0.0-x86_64-windows-msvc-release-mt.tar.xz
import json
import os
import subprocess
import sys
from pathlib import Path
from typing import Optional

import ccimport
from ccimport import compat, loader
from ccimport.core import get_full_file_name

from myclang.build_meta import ENABLE_JIT
from myclang.constants import MYCLANG_ROOT

def get_executable_path(executable: str) -> str:
    if compat.InWindows:
        cmd = ["powershell.exe", "(Get-Command {}).Path".format(executable)]
    else:
        cmd = ["which", executable]
    try:
        out = subprocess.check_output(cmd)
    except subprocess.CalledProcessError:
        return ""
    return out.decode("utf-8").strip()


def get_clang_root() -> Optional[Path]:
    clang_folder = os.getenv("LLVM_ROOT", None)
    if clang_folder:
        return Path(clang_folder)
    path = get_executable_path("clang")
    if path:
        clang_folder = Path(path).parent.parent / "lib"
    if clang_folder is None:
        return None
    return clang_folder.parent

flags = []
if not compat.InWindows:
    flags.append("-Wl,-rpath,{}".format(str(MYCLANG_ROOT)))

if ENABLE_JIT:
    with (Path(__file__).parent / "libclang.json").open("r") as f:
        LIBCLANG_BUILD_META_ALL = json.load(f)
    LIBCLANG_BUILD_META = LIBCLANG_BUILD_META_ALL[compat.OS.value]
    CLANG_ROOT = get_clang_root()
    msg = "you must provide clang root by either add clang++ to Path or export LLVM_ROOT"
    assert CLANG_ROOT is not None, msg
    LIBCLANG_NAME = "myclang"
    CLANG_COMPILER_NAME = "clangmain"
    CLANG_LIBPATH = CLANG_ROOT / "lib"

    if compat.InWindows:
        LIBCLANG_PATH = CLANG_ROOT / "bin" / "libclang.dll"
    LIBCLANG_INCLUDE = Path(__file__).parent.resolve() / "include"
    LIBCLANG_SOURCES = list((Path(__file__).parent / "libclang").glob("*.cpp"))
    CLANG_COMPILER_SOURCES = list((Path(__file__).parent / "clangcompiler").glob("*.cpp"))
    LIBCLANG_PATH = ccimport.ccimport(
        LIBCLANG_SOURCES,
        MYCLANG_ROOT / "myclang",
        includes=[CLANG_ROOT / "include", LIBCLANG_INCLUDE],
        libpaths=[CLANG_ROOT / "lib"],
        libraries=LIBCLANG_BUILD_META["libraries"],
        compile_options=LIBCLANG_BUILD_META["cflags"],
        link_options=LIBCLANG_BUILD_META["ldflags"],
        build_ctype=True,
        load_library=False,
        disable_hash=True)

    # FIXME find out SIGSEGV error when exit if import clang libraries
    clangutils = ccimport.autoimport([Path(__file__).parent / "clangutils.cc"],
                                     MYCLANG_ROOT / "clangutils",
                                     includes=[CLANG_ROOT / "include", LIBCLANG_INCLUDE],
                                     libpaths=[MYCLANG_ROOT],
                                     libraries=[LIBCLANG_NAME],
                                     link_options=flags)
    clcompiler = ccimport.autoimport([Path(__file__).parent / "clcompiler.cc"],
                                     MYCLANG_ROOT / "clcompiler",
                                     includes=[CLANG_ROOT / "include", LIBCLANG_INCLUDE],
                                     libpaths=[MYCLANG_ROOT],
                                     libraries=[LIBCLANG_NAME],
                                     link_options=flags)
else:
    LIBCLANG_PATH = Path(__file__).parent / get_full_file_name("myclang", True)
    clangutils = loader.try_import_from_path(
        Path(__file__).parent / get_full_file_name("clangutils", False))
    clcompiler = loader.try_import_from_path(
        Path(__file__).parent / get_full_file_name("clcompiler", False))



from myclang import astgen 
from pathlib import Path 
import pytest 
from ccimport import compat

_MSG = "vs 16.8 in github CI have a bug that cause this test fail, we need to wait for vs 16.9"
@pytest.mark.skipif(compat.InWindows, reason=_MSG)
def test_pch_gen():
    path = Path(__file__).parent.resolve() / "pch.h"
    astgen.create_pch(str(path), [], [])

@pytest.mark.skipif(compat.InWindows, reason=_MSG)
def test_from_ast_file():
    path = Path(__file__).parent.resolve() / "code.cc"
    astgen.from_ast_file([str(path)], [], [])




import subprocess
import time
from pathlib import Path
from typing import Deque, Dict, Iterator, List, Optional, Tuple, Union

import myclang.cindex as cindex
from myclang.cindex import Cursor, CursorKind, TranslationUnit

from ccimport import compat
from myclang.global_cfg import GLOBAL_CONFIG
from myclang.constants import MYCLANG_RESOURCE_INCLUDE, MYCLANG_CEXT_ROOT
from myclang.utils import tempdir
from myclang.resource import get_executable_path
import myclang
import os 
import sys


def create_pch(include: str,
               includes: List[str],
               options: List[str],
               cuda=False,
               compiler=None):
    """use pch can greatly increase ast gen speed for 
    header-only libraries.
    """
    # TODO: g++/msvc support (dont support cuda)
    cuda_flags: List[str] = []
    includes = includes.copy()
    includes.extend(GLOBAL_CONFIG.includes)
    includes = ["-I" + p for p in includes]
    if cuda:
        # see https://reviews.llvm.org/D87325 for '-S'
        cuda_flags = ["-x", "cuda", "-S", "--cuda-host-only"]
    pch_out_path = Path(include).parent / (Path(include).stem + ".pch")
    if compiler is None:
        compiler = myclang.get_clang_compiler_path()
        if compiler is None:
            raise ValueError("you need to provide clang compiler by download llvm or install myclang prebuilt")
    
    cmds = [
        str(compiler), *includes, *options, *cuda_flags, "-Xclang", "-emit-pch",
        include, "-o",
        str(pch_out_path)
    ]
    if compat.InWindows:
        # no way to add rpath-like to executable, so we need to append myclang.dll dir 
        env = os.environ.copy()
        paths = env["PATH"].split(";")
        paths.append(str(MYCLANG_CEXT_ROOT))
        env["PATH"] = ";".join(paths)
        return subprocess.check_call(cmds, env=env)

    return subprocess.check_call(cmds)


def from_ast_file(sources: List[Union[Path, str]],
                  includes: List[str],
                  options: List[str],
                  clang_path=None,
                  index: Optional[cindex.Index] = None
                  ) -> List[TranslationUnit]:
    """this function generate ast by subprocess ```clang++ -emit-ast```, 
    then read it by TranslationUnit.from_ast_file. 
    """
    # from myclang.cext import clcompiler

    source_paths: List[Path] = []
    includes = includes.copy()
    includes.extend(GLOBAL_CONFIG.includes)
    if compat.InWindows:
        assert get_executable_path("cl") is not None, "you must run inside vs develop prompt"
        # includes.extend(compat.get_system_include_paths("cl")), "you need to install vs in windows"
    elif compat.InLinux:
        assert get_executable_path("g++") is not None, "you need to install gcc in linux"
        includes.extend(compat.get_system_include_paths("g++"))
    else:
        raise NotImplementedError
    includes = ["-I" + str(p) for p in includes]
    for source_path in sources:
        source_path = Path(source_path)
        source_paths.extend(source_path.parent.glob(source_path.name))
    res: List[TranslationUnit] = []
    delete = True
    if clang_path is None:
        clang_path = myclang.get_clang_compiler_path()
        if clang_path is None:
            raise ValueError("you need to provide clang compiler by download llvm or install myclang prebuilt")

    if compat.InWindows:
        # you need to delete generated ast files by yourself.
        # because clang from_ast_file do something on ast file
        # and make it can't be deleted in current process.
        delete = False
    with tempdir(delete) as temp:
        cuda_flags = []
        for p in source_paths:
            if p.suffix == ".cu":
                # for cuda code, we only need host ast.
                # this can reduce compile time.
                cuda_flags.append("--cuda-host-only")
                break
        cmds = [
            str(clang_path), *includes, *options, *cuda_flags, "-emit-ast",
            *[str(p) for p in source_paths]
        ]
        # python_root = Path(sys.executable).parent.parent # xxx/bin/python -> xxx
        # correct_clang_root = myclang.get_fake_clang_root()
        # assert correct_clang_root is not None 
        # we replace MYCLANG_WRONG_PREFIX with MYCLANG_FIXED_ROOT
        # os.environ["MYCLANG_FIXED_ROOT"] = str(correct_clang_root)
        # os.environ["MYCLANG_WRONG_PREFIX"] = str(python_root)
        # TODO run clcompiler.compiler_main_bind(cmds) in subprocess to control .ast file
        # clcompiler.compiler_main_bind(cmds)
        if compat.InWindows:
            # no way to add rpath-like to executable, so we need to append myclang.dll dir 
            env = os.environ.copy()
            paths = env["PATH"].split(";")
            paths.append(str(MYCLANG_CEXT_ROOT))
            env["PATH"] = ";".join(paths)
            subprocess.check_call(cmds, cwd=str(temp), env=env, stderr=subprocess.STDOUT)
        else:
            subprocess.check_call(cmds, cwd=str(temp), stderr=subprocess.STDOUT)
        for source_path in source_paths:
            out_path: Path = temp / (source_path.stem + ".ast")
            res.append(TranslationUnit.from_ast_file(str(out_path), index))
    return res


def direct_gen(sources: List[Union[Path, str]],
               includes: List[str],
               options: List[str],
               parse_option=TranslationUnit.PARSE_NONE,
               index: Optional[cindex.Index] = None,
               unsaved_files: Optional[Tuple[Union[Path, str], str]] = None
               ) -> List[TranslationUnit]:
    """this function generate ast by TranslationUnit.from_source. 
    currently not working with cuda source. 
    """
    includes = includes.copy()
    includes.extend(GLOBAL_CONFIG.includes)
    includes.append(str(MYCLANG_RESOURCE_INCLUDE))
    if compat.InWindows:
        assert get_executable_path("cl") is not None, "you must run inside vs develop prompt"
        # includes.extend(compat.get_system_include_paths("cl")), "you need to install vs in windows"
    elif compat.InLinux:
        assert get_executable_path("g++") is not None, "you need to install gcc in linux"
        includes.extend(compat.get_system_include_paths("g++"))
    else:
        raise NotImplementedError
    source_paths: List[Path] = []
    includes = ["-I" + str(p) for p in includes]
    for source_path in sources:
        source_path = Path(source_path)
        source_paths.extend(source_path.parent.glob(source_path.name))
    res: List[TranslationUnit] = []

    for source_path in source_paths:
        if source_path.suffix == ".cu":
            raise NotImplementedError(
                "don't support cuda directly ast for now.")
        res.append(
            TranslationUnit.from_source(str(source_path),
                                        [*includes, *options],
                                        options=parse_option,
                                        index=index,
                                        unsaved_files=unsaved_files))
        for diag in res[-1].diagnostics:
            print(diag)

    return res


from .betterenums import NodeKind
from . import constants

def nkind_is_decl(nkind: NodeKind):
    K = nkind.value
    return ((K >= constants.CXCURSOR_FIRST_DECL and K <= constants.CXCURSOR_LAST_DECL) or
            (K >= constants.CXCURSOR_FIRST_EXTRA_DECL and K <= constants.CXCURSOR_LAST_EXTRA_DECL) or
            (K >= constants.CXCURSOREX_FIRST_DECL and K <= constants.CXCURSOREX_LAST_DECL))



from myclang.constants import MYCLANG_FAKE_CLANG_ROOT
from pathlib import Path 
import subprocess
from typing import List, Union 
import contextlib
from ccimport import compat

def get_executable_path(executable: str) -> str:
    if compat.InWindows:
        cmd = ["powershell.exe", "(Get-Command {}).Path".format(executable)]
    elif compat.InLinux:
        cmd = ["which", executable]
    else:
        raise NotImplementedError
    try:
        out = subprocess.check_output(cmd)
    except subprocess.CalledProcessError:
        return ""
    return out.decode("utf-8").strip()

def get_system_include(compiler: str="clang++") -> List[Path]:
    # TODO find a way to use clang headers instead of gcc/cl
    return compat.get_system_include_paths(compiler)


def get_fake_clang_root() -> Union[None, Path]:
    if MYCLANG_FAKE_CLANG_ROOT.exists():
        return MYCLANG_FAKE_CLANG_ROOT
    clang_exec = get_executable_path("clang++")
    if clang_exec is None:
        return None 
    return Path(clang_exec).parent.parent


def get_clang_compiler_path() -> Union[None, Path]:
    root = get_fake_clang_root()
    if not root:
        return None 
    return root / "bin" / "clang++"




from pathlib import Path

CODEAI_SAVE_ROOT = Path.home() / ".codeai"
CODEAI_SAVE_ROOT.mkdir(exist_ok=True)

MYCLANG_ROOT = CODEAI_SAVE_ROOT / "myclang"
MYCLANG_ROOT.mkdir(exist_ok=True)

MYCLANG_CEXT_ROOT = Path(__file__).parent / "cext"

MYCLANG_FAKE_CLANG_ROOT = Path(__file__).parent / "clang_fake_root"
MYCLANG_RESOURCE_INCLUDE = MYCLANG_FAKE_CLANG_ROOT / "lib" / "clang" / "11.0.0" / "include"

CXCURSOR_FIRST_DECL = 1
CXCURSOR_LAST_DECL = 39

CXCURSOR_FIRST_EXTRA_DECL = 600
CXCURSOR_LAST_EXTRA_DECL = 603

CXCURSOREX_FIRST_DECL = 1001
CXCURSOREX_LAST_DECL = 1003


import subprocess 
from ccimport import compat 
import contextlib 
import os 
import tempfile 
import shutil 
from pathlib import Path 
from myclang.constants import CODEAI_SAVE_ROOT




@contextlib.contextmanager
def tempdir(delete=True):
    if compat.InWindows:
        save_root = Path(CODEAI_SAVE_ROOT) / "codeai_tempdir"
        save_root.mkdir(exist_ok=True, parents=True)
        dirpath = save_root / os.urandom(24).hex()
        dirpath.mkdir()
        yield dirpath
        if delete:
            shutil.rmtree(dirpath)
    else:
        try:
            dirpath = tempfile.mkdtemp()
            dirpath = Path(dirpath)
            yield dirpath
        finally:
            if delete:
                shutil.rmtree(str(dirpath))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Note: To use the 'upload' functionality of this file, you must:
#   $ pip install twine

import io
import json
import os
import subprocess
import sys
from pathlib import Path
from shutil import rmtree
import shutil
from typing import Optional, Union

from ccimport import compat
from ccimport.extension import (AutoImportExtension, CCImportBuild,
                                CCImportExtension, ExtCallback)
from setuptools import Command, find_packages, setup

# Package meta-data.
NAME = 'myclang'
DESCRIPTION = 'standalone libclang code with some modifications'
URL = 'https://github.com/FindDefinition/myclang'
EMAIL = 'yanyan.sub@outlook.com'
AUTHOR = 'Yan Yan'
REQUIRES_PYTHON = '>=3.5'
VERSION = None

# What packages are required for this module to be executed?
REQUIRED = [
    "ccimport>=0.1.11",
]
if sys.version_info[:2] == (3, 6):
    REQUIRED.append("dataclasses")

# What packages are optional?
EXTRAS = {
    # 'fancy feature': ['django'],
}

# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!
# If you do change the License, remember to change the Trove Classifier for that!

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
    with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = '\n' + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

# Load the package's __version__.py module as a dictionary.
about = {}
if not VERSION:
    with open('version.txt', 'r') as f:
        version = f.read().strip()
else:
    version = VERSION
cwd = os.path.dirname(os.path.abspath(__file__))


def _convert_build_number(build_number):
    parts = build_number.split(".")
    if len(parts) == 2:
        return "{}{:03d}".format(int(parts[0]), int(parts[1]))
    elif len(parts) == 1:
        return build_number
    else:
        raise NotImplementedError


env_suffix = os.environ.get("CCIMPORT_VERSION_SUFFIX", "")
if env_suffix != "":
    version += ".dev{}".format(_convert_build_number(env_suffix))
version_path = os.path.join(cwd, NAME, '__version__.py')
about['__version__'] = version

with open(version_path, 'w') as f:
    f.write("__version__ = '{}'\n".format(version))
enable_jit = os.environ.get("MYCLANG_ENABLE_JIT", "1") == "1"
enable_jit_str = "True"
if not enable_jit:
    enable_jit_str = "False"
meta_path = os.path.join(cwd, NAME, 'build_meta.py')
with open(meta_path, 'w') as f:
    f.write("ENABLE_JIT = {}\n".format(enable_jit_str))

static_zlib_path = os.environ.get("MYCLANG_STATIC_ZLIB", None)
static_tinfo_path = os.environ.get("MYCLANG_STATIC_TINFO", None)


class UploadCommand(Command):
    """Support setup.py upload."""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds...')
            rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution...')
        os.system('{0} setup.py sdist bdist_wheel --universal'.format(
            sys.executable))

        self.status('Uploading the package to PyPI via Twine...')
        os.system('twine upload dist/*')

        self.status('Pushing git tags...')
        os.system('git tag v{0}'.format(about['__version__']))
        os.system('git push --tags')

        sys.exit()

class CreateLinkCallback(ExtCallback):
    def __call__(self, ext: Union["AutoImportExtension", "CCImportExtension"], extdir: Path, target_path: Path):
        clang_name = "clang"
        clangpp_name = "clang++"
        if compat.InWindows:
            clang_name = "clang.exe"
            clangpp_name = "clang++.exe"

        assert target_path.name == clang_name
        clangpp_p = extdir / "myclang" / "clang_fake_root" / "bin" / clangpp_name
        shutil.copy(str(target_path), str(clangpp_p))
        # permission denied when create symlink in windows. 
        # so we just copy them.

        # clangpp_p.symlink_to(target_path, False)


def get_executable_path(executable: str) -> str:
    if compat.InWindows:
        cmd = ["powershell.exe", "(Get-Command {}).Path".format(executable)]
    else:
        cmd = ["which", executable]
    try:
        out = subprocess.check_output(cmd)
    except subprocess.CalledProcessError:
        return ""
    return out.decode("utf-8").strip()


def get_clang_root() -> Optional[Path]:
    clang_folder = os.getenv("LLVM_ROOT", None)
    if clang_folder:
        return Path(clang_folder)
    path = get_executable_path("clang")
    if path:
        clang_folder = Path(path).parent.parent / "lib"
    if clang_folder is None:
        return None
    return clang_folder.parent


if enable_jit:
    cmdclass = {
        'upload': UploadCommand,
    }
    ext_modules = []
else:
    LIBCLANG_MODULE_PATH = Path(__file__).parent / "myclang" / "cext"
    with (LIBCLANG_MODULE_PATH / "libclang.json").open("r") as f:
        LIBCLANG_BUILD_META_ALL = json.load(f)
    LIBCLANG_BUILD_META = LIBCLANG_BUILD_META_ALL[compat.OS.value]
    if 'tinfo' in LIBCLANG_BUILD_META["libraries"]:
        # currently tinfo static build have bug in manylinux docker
        LIBCLANG_BUILD_META["libraries"].remove('tinfo')

    if static_zlib_path is not None and "z" in LIBCLANG_BUILD_META["libraries"]:
        idx = LIBCLANG_BUILD_META["libraries"].index("z")
        if Path(static_zlib_path).is_absolute():
            LIBCLANG_BUILD_META["libraries"][idx] = "path::" + static_zlib_path
        else:
            LIBCLANG_BUILD_META["libraries"][idx] = "file::" + static_zlib_path
    if static_tinfo_path is not None and "tinfo" in LIBCLANG_BUILD_META["libraries"]:
        idx = LIBCLANG_BUILD_META["libraries"].index("tinfo")
        if Path(static_tinfo_path).is_absolute():
            LIBCLANG_BUILD_META["libraries"][idx] = "path::" + static_tinfo_path
        else:
            LIBCLANG_BUILD_META["libraries"][idx] = "file::" + static_tinfo_path

    CLANG_ROOT = get_clang_root()
    assert CLANG_ROOT is not None, "can't find clang, install clang first."
    cmdclass = {
        'upload': UploadCommand,
        'build_ext': CCImportBuild,
    }
    LIBCLANG_INCLUDE = LIBCLANG_MODULE_PATH.resolve() / "include"

    LIBCLANG_SOURCES = list((LIBCLANG_MODULE_PATH / "libclang").glob("*.cpp"))
    libclang_ext = CCImportExtension(
        "myclang",
        LIBCLANG_SOURCES,
        "myclang/cext/myclang",
        includes=[CLANG_ROOT / "include", LIBCLANG_INCLUDE],
        libpaths=[CLANG_ROOT / "lib"],
        libraries=LIBCLANG_BUILD_META["libraries"],
        compile_options=LIBCLANG_BUILD_META["cflags"],
        link_options=LIBCLANG_BUILD_META["ldflags"],
        build_ctype=True,
        std="c++14",
    )
    flags = []
    clang_main_flags = []
    if not compat.InWindows:
        flags.append("-Wl,--enable-new-dtags")
        # ninja need to escape dollar sign
        flags.append("-Wl,-rpath='$$ORIGIN'")
        clang_main_flags.append("-Wl,--enable-new-dtags")
        clang_main_flags.append("-Wl,-rpath='$$ORIGIN/../../cext'")

    clangutils_ext = AutoImportExtension(
        "clangutils",
        [LIBCLANG_MODULE_PATH / "clangutils.cc"],
        "myclang/cext/clangutils",
        includes=[CLANG_ROOT / "include", LIBCLANG_INCLUDE],
        libpaths=["{extdir}/myclang/cext"],
        libraries=["myclang"],
        link_options=flags,
        std="c++14",
    )
    clcompiler_ext = AutoImportExtension(
        "clcompiler",
        [LIBCLANG_MODULE_PATH / "clcompiler.cc"],
        "myclang/cext/clcompiler",
        includes=[CLANG_ROOT / "include", LIBCLANG_INCLUDE],
        libpaths=["{extdir}/myclang/cext"],
        libraries=["myclang"],
        link_options=flags,
        std="c++14",
    )
    compiler_path = "myclang/clang_fake_root/bin/clang"
    clang_compiler_ext = CCImportExtension(
        "clang",
        [LIBCLANG_MODULE_PATH / "clangmain.cc"],
        compiler_path,
        includes=[CLANG_ROOT / "include", LIBCLANG_INCLUDE],
        libpaths=["{extdir}/myclang/cext"],
        libraries=["myclang"],
        link_options=clang_main_flags,
        std="c++14",
        shared=False, 
        extcallback=CreateLinkCallback(),
    )

    ext_modules = [libclang_ext, clangutils_ext, clcompiler_ext, clang_compiler_ext]

# Where the magic happens:
setup(
    name=NAME,
    version=about['__version__'],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(exclude=('tests', )),
    # If your package is a single module, use this instead of 'packages':
    # py_modules=['mypackage'],
    entry_points={
        'console_scripts': [],
    },
    install_requires=REQUIRED,
    extras_require=EXTRAS,
    include_package_data=True,
    license='MIT',
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: Implementation :: CPython',
    ],
    # $ setup.py publish support.
    cmdclass=cmdclass,
    ext_modules=ext_modules,
)



import json
import traceback
from pathlib import Path
from typing import List

from myclang.constants import CODEAI_SAVE_ROOT


class GlobalConfig:
    def __init__(self, includes: List[str]):
        self.includes = includes


GLOBAL_CONFIG_PATH = CODEAI_SAVE_ROOT / "myclang_global.json"
GLOBAL_CONFIG = GlobalConfig([])

if GLOBAL_CONFIG_PATH.exists():
    try:
        with GLOBAL_CONFIG_PATH.open("r") as f:
            cfg = json.load(f)
        GLOBAL_CONFIG = GlobalConfig(cfg["includes"])
    except:
        traceback.print_exc()



"""copy from codeai.core.dump.tree
"""
from typing import Generic, TypeVar, Any, Callable, List, Iterable
import io

_T = TypeVar("_T")


class TreeDumpStatus(Generic[_T]):
    def __init__(self, node: _T, indent: int, depth: int, direct_prefix: str,
                 child_prefixes: List[str]):
        self.node = node
        self.indent = indent
        self.depth = depth
        # node prefix: "".join(child_prefixes[:-1]) + direct_prefix
        self.direct_prefix = direct_prefix
        self.child_prefixes = child_prefixes

    def get_prefix(self):
        return "".join(self.child_prefixes[:-1]) + self.direct_prefix


def tree_dumps(node: _T,
               printer: Callable[[_T], str],
               getchild: Callable[[_T], Iterable[_T]],
               indent: int = 2,
               start_indent=0,
               max_depth=9999,
               tree_formatter=lambda x: x) -> str:
    ss = io.StringIO()
    q = [TreeDumpStatus(node, start_indent, 0, "",
                        [])]  # type: List[TreeDumpStatus]
    while q:
        status = q.pop()
        cur_node = status.node
        ind = status.indent
        if status.depth >= max_depth:
            continue
        cu_str = printer(cur_node)
        print(tree_formatter(status.get_prefix()) + cu_str, file=ss)
        nexts = []
        childs = list(getchild(cur_node))
        for i, c in enumerate(childs):
            direct_prefix = "|-"
            child_prefixes = status.child_prefixes + ["| "]
            if i == len(childs) - 1:
                direct_prefix = "`-"
                child_prefixes = status.child_prefixes + ["  "]
            next_status = TreeDumpStatus(c, ind + indent, status.depth + 1,
                                         direct_prefix, child_prefixes)
            nexts.append(next_status)
        q.extend(nexts[::-1])
    return ss.getvalue()
